@@ -18,6 +18,8 @@ static bool		have_little_core;
 const volatile u16	cpu_order_performance[LAVD_CPU_ID_MAX]; /* CPU preference order for performance and balanced mode */
 const volatile u16	cpu_order_powersave[LAVD_CPU_ID_MAX]; /* CPU preference order for powersave mode */
 const volatile u16	cpu_capacity[LAVD_CPU_ID_MAX]; /* CPU capacity based on 1024 */
+const volatile u8	cpu_big[LAVD_CPU_ID_MAX]; /* Is a CPU a big core? */
+const volatile u8	cpu_turbo[LAVD_CPU_ID_MAX]; /* Is a CPU a turbo core? */
 
 static int		nr_cpdoms; /* number of compute domains */
 struct cpdom_ctx	cpdom_ctxs[LAVD_CPDOM_MAX_NR]; /* contexts for compute domains */
@@ -127,10 +129,10 @@ static void do_core_compaction(void)
 {
 	const volatile u16 *cpu_order = get_cpu_order();
 	struct cpu_ctx *cpuc;
-	struct bpf_cpumask *active, *ovrflw, *cd_cpumask;
+	struct bpf_cpumask *active, *ovrflw;
 	struct cpdom_ctx *cpdomc;
 	int nr_active, nr_active_old, cpu, i;
-	u32 sum_capacity = 0, big_capacity = 0;
+	u32 sum_capacity = 0, big_capacity = 0, nr_active_cpdoms = 0;
 	bool clear;
 	u64 cpdom_id;
 
@@ -172,12 +174,18 @@ static void do_core_compaction(void)
 		if (i < nr_active) {
 			bpf_cpumask_set_cpu(cpu, active);
 			bpf_cpumask_clear_cpu(cpu, ovrflw);
-
-			cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpu]);
-			if (cpdomc)
-				WRITE_ONCE(cpdomc->is_active, true);
 			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
 
+			/*
+			 * Accumulate the capacity of active CPUs and
+			 * increase the number of active CPUs.
+			 */
+			cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpuc->cpdom_id]);
+			if (cpdomc) {
+				cpdomc->cap_sum_temp += cpuc->capacity;
+				cpdomc->nr_acpus_temp++;
+			}
+
 			/*
 			 * Calculate big capacity ratio among active cores.
 			 */
@@ -213,20 +221,24 @@ static void do_core_compaction(void)
 	sys_stat.nr_active = nr_active;
 
 	/*
-	 * Maintain cpdomc->is_active reflecting the active set.
+	 * Update nr_active_cpus and cap_sum_active_cpus.
 	 */
 	bpf_for(cpdom_id, 0, nr_cpdoms) {
 		if (cpdom_id >= LAVD_CPDOM_MAX_NR)
 			break;
 
 		cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpdom_id]);
-		cd_cpumask = MEMBER_VPTR(cpdom_cpumask, [cpdom_id]);
-		if (!cpdomc || !cd_cpumask || !cpdomc->is_active)
+		if (!cpdomc)
 			continue;
+		WRITE_ONCE(cpdomc->nr_active_cpus, cpdomc->nr_acpus_temp);
+		WRITE_ONCE(cpdomc->nr_acpus_temp, 0);
+		WRITE_ONCE(cpdomc->cap_sum_active_cpus, cpdomc->cap_sum_temp);
+		WRITE_ONCE(cpdomc->cap_sum_temp, 0);
 
-		if (!bpf_cpumask_intersects(cast_mask(active), cast_mask(cd_cpumask)))
-			WRITE_ONCE(cpdomc->is_active, false);
+		if (cpdomc->nr_active_cpus)
+			nr_active_cpdoms++;
 	}
+	sys_stat.nr_active_cpdoms = nr_active_cpdoms;
 
 unlock_out:
 	bpf_rcu_read_unlock();
@@ -417,6 +429,9 @@ static int reinit_active_cpumask_for_performance(void)
 	struct cpu_ctx *cpuc;
 	struct bpf_cpumask *active, *ovrflw;
 	const struct cpumask *online_cpumask;
+	struct cpdom_ctx *cpdomc;
+	u64 dsq_id;
+	u32 nr_active_cpdoms = 0;
 	int cpu, err = 0;
 
 	barrier();
@@ -444,19 +459,28 @@ static int reinit_active_cpumask_for_performance(void)
 	if (have_little_core) {
 		bpf_for(cpu, 0, nr_cpu_ids) {
 			cpuc = get_cpu_ctx_id(cpu);
-			if (!cpuc) {
-				err = -ESRCH;
-				goto unlock_out;
+			if (!cpuc)
+				continue;
+			if (!cpuc->is_online) {
+				bpf_cpumask_clear_cpu(cpu, active);
+				bpf_cpumask_clear_cpu(cpu, ovrflw);
+				continue;
 			}
 
 			if (cpuc->big_core) {
 				bpf_cpumask_set_cpu(cpu, active);
 				bpf_cpumask_clear_cpu(cpu, ovrflw);
-			}
-			else {
+			} else {
 				bpf_cpumask_set_cpu(cpu, ovrflw);
 				bpf_cpumask_clear_cpu(cpu, active);
 			}
+			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+
+			cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpuc->cpdom_id]);
+			if (cpdomc) {
+				cpdomc->nr_acpus_temp++;
+				cpdomc->cap_sum_temp += cpuc->capacity;
+			}
 		}
 	} else {
 		online_cpumask = scx_bpf_get_online_cpumask();
@@ -465,8 +489,41 @@ static int reinit_active_cpumask_for_performance(void)
 		scx_bpf_put_cpumask(online_cpumask);
 
 		bpf_cpumask_clear(ovrflw);
+
+		bpf_for(cpu, 0, nr_cpu_ids) {
+			cpuc = get_cpu_ctx_id(cpu);
+			if (!cpuc || !cpuc->is_online)
+				continue;
+
+			scx_bpf_kick_cpu(cpu, SCX_KICK_IDLE);
+
+			cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpuc->cpdom_id]);
+			if (cpdomc) {
+				cpdomc->nr_acpus_temp++;
+				cpdomc->cap_sum_temp += cpuc->capacity;
+			}
+		}
+
+	}
+
+	/*
+	 * Update nr_active_cpus and cap_sum_active_cpus.
+	 */
+	bpf_for(dsq_id, 0, nr_cpdoms) {
+		if (dsq_id >= LAVD_CPDOM_MAX_NR)
+			break;
+
+		cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
+		WRITE_ONCE(cpdomc->nr_active_cpus, cpdomc->nr_acpus_temp);
+		WRITE_ONCE(cpdomc->nr_acpus_temp, 0);
+		WRITE_ONCE(cpdomc->cap_sum_active_cpus, cpdomc->cap_sum_temp);
+		WRITE_ONCE(cpdomc->cap_sum_temp, 0);
+
+		if (cpdomc->nr_active_cpus)
+			nr_active_cpdoms++;
 	}
 	sys_stat.nr_active = nr_cpus_onln;
+	sys_stat.nr_active_cpdoms = nr_active_cpdoms;
 
 unlock_out:
 	bpf_rcu_read_unlock();
@@ -475,7 +532,7 @@ unlock_out:
 
 static void update_cpuperf_target(struct cpu_ctx *cpuc)
 {
-	u32 util, cpuperf_target;
+	u32 util, max_util, cpuperf_target;
 
 	/*
 	 * The CPU utilization decides the frequency. The bigger one between
@@ -484,8 +541,9 @@ static void update_cpuperf_target(struct cpu_ctx *cpuc)
 	 * LAVD_CPU_UTIL_MAX_FOR_CPUPERF (85%), ceil to 100%.
 	 */
 	if (!no_freq_scaling) {
-		util = max(cpuc->avg_util, cpuc->cur_util) <
-			LAVD_CPU_UTIL_MAX_FOR_CPUPERF? : LAVD_SCALE;
+		max_util = max(cpuc->avg_util, cpuc->cur_util);
+		util = (max_util < LAVD_CPU_UTIL_MAX_FOR_CPUPERF) ? max_util
+								  : LAVD_SCALE;
 		cpuperf_target = (util * SCX_CPUPERF_ONE) >> LAVD_SHIFT;
 	} else
 		cpuperf_target = SCX_CPUPERF_ONE;
@@ -508,37 +566,16 @@ static void reset_cpuperf_target(struct cpu_ctx *cpuc)
 
 static u16 get_cpuperf_cap(s32 cpu)
 {
-	if (cpu >= 0 && cpu < nr_cpu_ids && cpu < LAVD_CPU_ID_MAX)
-		return cpu_capacity[cpu];
+	const volatile u16 *cap;
+
+	cap = MEMBER_VPTR(cpu_capacity, [cpu]);
+	if (cap)
+		return *cap;
 
 	debugln("Infeasible CPU id: %d", cpu);
 	return 0;
 }
 
-static u16 get_cputurbo_cap(void)
-{
-	u16 turbo_cap = 0;
-	int nr_turbo = 0, cpu;
-
-	/*
-	 * Find the maximum CPU capacity
-	 */
-	for (cpu = 0; cpu < nr_cpu_ids && cpu < LAVD_CPU_ID_MAX; cpu++) {
-		if (cpu_capacity[cpu] > turbo_cap) {
-			turbo_cap = cpu_capacity[cpu];
-			nr_turbo++;
-		}
-	}
-
-	/*
-	 * If all CPU's capacities are the same, ignore the turbo.
-	 */
-	if (nr_turbo <= 1)
-		turbo_cap = 0;
-
-	return turbo_cap;
-}
-
 static u64 scale_cap_freq(u64 dur, s32 cpu)
 {
 	u64 cap, freq, scaled_dur;
