@@ -39,7 +39,6 @@ struct sys_stat_ctx {
 	u32		nr_perf_cri;
 	u32		nr_lat_cri;
 	u32		nr_x_migration;
-	u32		nr_stealee;
 	u32		nr_big;
 	u32		nr_pc_on_big;
 	u32		nr_lc_on_big;
@@ -62,67 +61,28 @@ static void init_sys_stat_ctx(struct sys_stat_ctx *c)
 	sys_stat.last_update_clk = c->now;
 }
 
-static void plan_x_cpdom_migration(struct sys_stat_ctx *c)
+static void collect_sys_stat(struct sys_stat_ctx *c)
 {
 	struct cpdom_ctx *cpdomc;
 	u64 dsq_id;
-	u32 avg_nr_q_tasks_per_cpu = 0, nr_q_tasks, x_mig_delta;
-	u32 stealer_threshold, stealee_threshold;
-
-	/*
-	 * Calcualte average queued tasks per CPU per compute domain.
-	 */
-	bpf_for(dsq_id, 0, nr_cpdoms) {
-		if (dsq_id >= LAVD_CPDOM_MAX_NR)
-			break;
-
-		nr_q_tasks = scx_bpf_dsq_nr_queued(dsq_id);
-		c->nr_queued_task += nr_q_tasks;
-
-		cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
-		cpdomc->nr_q_tasks_per_cpu = (nr_q_tasks << LAVD_SHIFT) / cpdomc->nr_cpus;
-		avg_nr_q_tasks_per_cpu += cpdomc->nr_q_tasks_per_cpu;
-	}
-	avg_nr_q_tasks_per_cpu /= nr_cpdoms;
-
-	/*
-	 * Determine stealer and stealee domains.
-	 *
-	 * A stealer domain, whose per-CPU queue length is shorter than
-	 * the average, will steal a task from any of stealee domain,
-	 * whose per-CPU queue length is longer than the average.
-	 * Compute domain around average will not do anything.
-	 */
-	x_mig_delta = avg_nr_q_tasks_per_cpu >> LAVD_CPDOM_MIGRATION_SHIFT;
-	stealer_threshold = avg_nr_q_tasks_per_cpu - x_mig_delta;
-	stealee_threshold = avg_nr_q_tasks_per_cpu + x_mig_delta;
-
-	bpf_for(dsq_id, 0, nr_cpdoms) {
-		if (dsq_id >= LAVD_CPDOM_MAX_NR)
-			break;
-
-		cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
-
-		if (cpdomc->nr_q_tasks_per_cpu < stealer_threshold) {
-			WRITE_ONCE(cpdomc->is_stealer, true);
-			WRITE_ONCE(cpdomc->is_stealee, false);
-		}
-		else if (cpdomc->nr_q_tasks_per_cpu > stealee_threshold) {
-			WRITE_ONCE(cpdomc->is_stealer, false);
-			WRITE_ONCE(cpdomc->is_stealee, true);
-			c->nr_stealee++;
-		}
-		else {
-			WRITE_ONCE(cpdomc->is_stealer, false);
-			WRITE_ONCE(cpdomc->is_stealee, false);
-		}
-	}
-}
-
-static void collect_sys_stat(struct sys_stat_ctx *c)
-{
 	int cpu;
 
+	/*
+	 * Collect statistics for each compute domain.
+	 */
+	bpf_for(dsq_id, 0, nr_cpdoms) {
+		if (dsq_id >= LAVD_CPDOM_MAX_NR)
+			break;
+
+		cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
+		cpdomc->cur_util_sum = 0;
+		cpdomc->nr_queued_task = scx_bpf_dsq_nr_queued(dsq_id);
+		c->nr_queued_task += cpdomc->nr_queued_task;
+	}
+
+	/*
+	 * Collect statistics for each CPU.
+	 */
 	bpf_for(cpu, 0, nr_cpu_ids) {
 		struct cpu_ctx *cpuc = get_cpu_ctx_id(cpu);
 		if (!cpuc) {
@@ -221,7 +181,7 @@ static void collect_sys_stat(struct sys_stat_ctx *c)
 		}
 
 		/*
-		 * Calculcate per-CPU utilization
+		 * Calculcate per-CPU utilization.
 		 */
 		u64 compute = 0;
 		if (c->duration > cpuc->idle_total)
@@ -230,8 +190,12 @@ static void collect_sys_stat(struct sys_stat_ctx *c)
 		cpuc->cur_util = (compute << LAVD_SHIFT) / c->duration;
 		cpuc->avg_util = calc_asym_avg(cpuc->avg_util, cpuc->cur_util);
 
+		cpdomc = MEMBER_VPTR(cpdom_ctxs, [cpuc->cpdom_id]);
+		if (cpdomc)
+			cpdomc->cur_util_sum += cpuc->cur_util;
+
 		/*
-		 * Accmulate system-wide idle time
+		 * Accmulate system-wide idle time.
 		 */
 		c->idle_total += cpuc->idle_total;
 		cpuc->idle_total = 0;
@@ -288,7 +252,7 @@ static void calc_sys_stat(struct sys_stat_ctx *c)
 	sys_stat.max_lat_cri = calc_avg32(sys_stat.max_lat_cri, c->max_lat_cri);
 	sys_stat.avg_lat_cri = calc_avg32(sys_stat.avg_lat_cri, c->avg_lat_cri);
 	sys_stat.thr_lat_cri = sys_stat.max_lat_cri - ((sys_stat.max_lat_cri -
-				sys_stat.avg_lat_cri) >> LAVD_LC_PREEMPT_SHIFT);
+				sys_stat.avg_lat_cri) >> preempt_shift);
 
 	if (have_little_core) {
 		sys_stat.min_perf_cri =
@@ -299,8 +263,6 @@ static void calc_sys_stat(struct sys_stat_ctx *c)
 			calc_avg32(sys_stat.max_perf_cri, c->max_perf_cri);
 	}
 
-	sys_stat.nr_stealee = c->nr_stealee;
-
 	if (c->nr_sched > 0)
 		avg_svc_time = c->tot_svc_time / c->nr_sched;
 	sys_stat.avg_svc_time = calc_avg(sys_stat.avg_svc_time, avg_svc_time);
@@ -359,28 +321,46 @@ static void do_update_sys_stat(void)
 	struct sys_stat_ctx c;
 
 	init_sys_stat_ctx(&c);
-	plan_x_cpdom_migration(&c);
 	collect_sys_stat(&c);
 	calc_sys_stat(&c);
 }
 
 static void update_sys_stat(void)
 {
+	/*
+	 * Update system statistics.
+	 */
 	do_update_sys_stat();
 
+	/*
+	 * Change the power profile based on the statistics.
+	 */
 	if (is_autopilot_on)
 		do_autopilot();
 
+	/*
+	 * Perform core compaction for powersave and balance mode.
+	 * Or turn on all CPUs for performance mode.
+	 */
 	if (!no_core_compaction)
 		do_core_compaction();
 
-	calc_sys_time_slice();
-	update_thr_perf_cri();
-
 	if (reinit_cpumask_for_performance) {
 		reinit_cpumask_for_performance = false;
 		reinit_active_cpumask_for_performance();
 	}
+
+	/*
+	 * Update time slice and performance criticality threshold.
+	 */
+	calc_sys_time_slice();
+	update_thr_perf_cri();
+
+	/*
+	 * Plan cross-domain task migration.
+	 */
+	if (nr_cpdoms > 1)
+		plan_x_cpdom_migration();
 }
 
 static int update_timer_cb(void *map, int *key, struct bpf_timer *timer)
@@ -398,12 +378,23 @@ static int update_timer_cb(void *map, int *key, struct bpf_timer *timer)
 
 static s32 init_sys_stat(u64 now)
 {
+	struct cpdom_ctx *cpdomc;
 	struct bpf_timer *timer;
+	u64 dsq_id;
 	u32 key = 0;
 	int err;
 
 	sys_stat.last_update_clk = now;
 	sys_stat.nr_active = nr_cpus_onln;
+	sys_stat.slice = LAVD_SLICE_MAX_NS_DFL;
+	bpf_for(dsq_id, 0, nr_cpdoms) {
+		if (dsq_id >= LAVD_CPDOM_MAX_NR)
+			break;
+
+		cpdomc = MEMBER_VPTR(cpdom_ctxs, [dsq_id]);
+		if (cpdomc->nr_active_cpus)
+			sys_stat.nr_active_cpdoms++;
+	}
 
 	timer = bpf_map_lookup_elem(&update_timer, &key);
 	if (!timer) {
