@@ -30,6 +30,7 @@ use std::time::Duration;
 use anyhow::Context;
 use anyhow::Result;
 use clap::Parser;
+use clap_num::number_range;
 use crossbeam::channel;
 use crossbeam::channel::Receiver;
 use crossbeam::channel::RecvTimeoutError;
@@ -53,6 +54,7 @@ use scx_utils::scx_ops_open;
 use scx_utils::set_rlimit_infinity;
 use scx_utils::uei_exited;
 use scx_utils::uei_report;
+use scx_utils::CoreType;
 use scx_utils::Cpumask;
 use scx_utils::EnergyModel;
 use scx_utils::Topology;
@@ -115,9 +117,16 @@ struct Opts {
     slice_max_us: u64,
 
     /// Minimum scheduling slice duration in microseconds.
-    #[clap(long = "slice-min-us", default_value = "300")]
+    #[clap(long = "slice-min-us", default_value = "500")]
     slice_min_us: u64,
 
+    /// Limit the ratio of preemption to the roughly top P% of latency-critical
+    /// tasks. When N is given as an argument, P is 0.5^N * 100. The default
+    /// value is 6, which limits the preemption for the top 1.56% of
+    /// latency-critical tasks.
+    #[clap(long = "preempt-shift", default_value = "6", value_parser=Opts::preempt_shift_range)]
+    preempt_shift: u8,
+
     /// List of CPUs in preferred order (e.g., "0-3,7,6,5,4"). The scheduler
     /// uses the CPU preference mode only when the core compaction is enabled
     /// (i.e., balanced or powersave mode is specified as an option or chosen
@@ -273,6 +282,10 @@ impl Opts {
 
         Some(self)
     }
+
+    fn preempt_shift_range(s: &str) -> Result<u8, String> {
+        number_range(s, 0, 10)
+    }
 }
 
 unsafe impl Plain for msg_task_ctx {}
@@ -301,6 +314,8 @@ struct CpuFlatId {
     smt_level: usize,
     cache_size: usize,
     cpu_cap: usize,
+    big_core: bool,
+    turbo_core: bool,
 }
 
 #[derive(Debug, Eq, PartialEq, Ord, PartialOrd, Clone)]
@@ -351,13 +366,12 @@ impl FlatTopology {
         debug!("{:#?}", sys_topo);
         debug!("{:#?}", sys_em);
 
-        let (cpu_fids_performance, avg_cap) =
-            Self::build_cpu_fids(&sys_topo, &sys_em, false).unwrap();
-        let (cpu_fids_powersave, _) = Self::build_cpu_fids(&sys_topo, &sys_em, true).unwrap();
+        let cpu_fids_performance = Self::build_cpu_fids(&sys_topo, &sys_em, false).unwrap();
+        let cpu_fids_powersave = Self::build_cpu_fids(&sys_topo, &sys_em, true).unwrap();
 
         // Note that building compute domain is not dependent to CPU orer
         // so it is okay to use any cpu_fids_*.
-        let cpdom_map = Self::build_cpdom(&cpu_fids_performance, avg_cap).unwrap();
+        let cpdom_map = Self::build_cpdom(&cpu_fids_performance).unwrap();
 
         Ok(FlatTopology {
             all_cpus_mask: sys_topo.span,
@@ -373,11 +387,10 @@ impl FlatTopology {
         topo: &Topology,
         em: &Result<EnergyModel>,
         prefer_powersave: bool,
-    ) -> Option<(Vec<CpuFlatId>, usize)> {
+    ) -> Option<Vec<CpuFlatId>> {
         let mut cpu_fids = Vec::new();
 
         // Build a vector of cpu flat ids.
-        let mut avg_cap = 0;
         for (&node_id, node) in topo.nodes.iter() {
             for (llc_pos, (_llc_id, llc)) in node.llcs.iter().enumerate() {
                 for (core_pos, (_core_id, core)) in llc.cores.iter().enumerate() {
@@ -394,14 +407,14 @@ impl FlatTopology {
                             smt_level: cpu.smt_level,
                             cache_size: cpu.cache_size,
                             cpu_cap: cpu.cpu_capacity,
+                            big_core: cpu.core_type != CoreType::Little,
+                            turbo_core: cpu.core_type == CoreType::Big { turbo: true },
                         };
                         cpu_fids.push(RefCell::new(cpu_fid));
-                        avg_cap += cpu.cpu_capacity;
                     }
                 }
             }
         }
-        avg_cap /= cpu_fids.len() as usize;
 
         // Convert a vector of RefCell to a vector of plain cpu_fids
         let mut cpu_fids2 = Vec::new();
@@ -427,13 +440,14 @@ impl FlatTopology {
                 });
             }
             false => {
-                // Sort the cpu_fids by cpu, node, llc, ^cpu_cap, smt_level, ^cache_size, perf_dom, and core order
+                // Sort the cpu_fids by node, llc, ^cpu_cap, cpu_pos, smt_level, ^cache_size, perf_dom, and core order
+                // For performance mode, prioritize CPU capacity over physical position for ARM big.LITTLE systems
                 cpu_fids.sort_by(|a, b| {
-                    a.cpu_pos
-                        .cmp(&b.cpu_pos)
-                        .then_with(|| a.node_id.cmp(&b.node_id))
-                        .then_with(|| a.llc_pos.cmp(&b.llc_pos))
-                        .then_with(|| b.cpu_cap.cmp(&a.cpu_cap))
+                    a.node_id
+                        .cmp(&b.node_id) // NUMA node first
+                        .then_with(|| a.llc_pos.cmp(&b.llc_pos)) // LLC locality
+                        .then_with(|| b.cpu_cap.cmp(&a.cpu_cap)) // CPU performance first (^cpu_cap)
+                        .then_with(|| a.cpu_pos.cmp(&b.cpu_pos)) // Physical position as tie-breaker
                         .then_with(|| a.smt_level.cmp(&b.smt_level))
                         .then_with(|| b.cache_size.cmp(&a.cache_size))
                         .then_with(|| a.pd_id.cmp(&b.pd_id))
@@ -442,7 +456,7 @@ impl FlatTopology {
             }
         }
 
-        Some((cpu_fids, avg_cap))
+        Some(cpu_fids)
     }
 
     /// Get the performance domain (i.e., CPU frequency domain) ID for a CPU.
@@ -457,17 +471,17 @@ impl FlatTopology {
     /// Build a list of compute domains
     fn build_cpdom(
         cpu_fids: &Vec<CpuFlatId>,
-        avg_cap: usize,
     ) -> Option<BTreeMap<ComputeDomainKey, ComputeDomainValue>> {
         // Creat a compute domain map, where a compute domain is a CPUs that
         // are under the same node and LLC and have the same core type.
         let mut cpdom_id = 0;
         let mut cpdom_map: BTreeMap<ComputeDomainKey, ComputeDomainValue> = BTreeMap::new();
+        let mut cpdom_types: BTreeMap<usize, bool> = BTreeMap::new();
         for cpu_fid in cpu_fids.iter() {
             let key = ComputeDomainKey {
                 node_id: cpu_fid.node_id,
                 llc_pos: cpu_fid.llc_pos,
-                is_big: cpu_fid.cpu_cap >= avg_cap,
+                is_big: cpu_fid.big_core,
             };
             let mut value;
             match cpdom_map.get(&key) {
@@ -481,6 +495,8 @@ impl FlatTopology {
                         cpu_ids: Vec::new(),
                         neighbor_map: RefCell::new(BTreeMap::new()),
                     };
+                    cpdom_types.insert(cpdom_id, key.is_big);
+
                     cpdom_id += 1;
                 }
             }
@@ -488,18 +504,6 @@ impl FlatTopology {
             cpdom_map.insert(key, value);
         }
 
-        // Fill up cpdom_alt_id for each compute domain, where the alternative
-        // compute domain is a compute domain that are under the same node
-        // and LLC but has a different core type.
-        for (k, v) in cpdom_map.iter() {
-            let mut key = k.clone();
-            key.is_big = !k.is_big;
-
-            if let Some(alt_v) = cpdom_map.get(&key) {
-                v.cpdom_alt_id.set(alt_v.cpdom_id);
-            }
-        }
-
         // Build a neighbor map for each compute domain, where neighbors are
         // ordered by core type, node, and LLC.
         for ((from_k, from_v), (to_k, to_v)) in iproduct!(cpdom_map.iter(), cpdom_map.iter()) {
@@ -519,6 +523,35 @@ impl FlatTopology {
             }
         }
 
+        // Fill up cpdom_alt_id for each compute domain.
+        for (k, v) in cpdom_map.iter() {
+            let mut key = k.clone();
+            key.is_big = !k.is_big;
+
+            if let Some(alt_v) = cpdom_map.get(&key) {
+                // First, try to find an alternative domain
+                // under the same node/LLC.
+                v.cpdom_alt_id.set(alt_v.cpdom_id);
+            } else {
+                // If there is no alternative domain in the same node/LLC,
+                // choose the closest one.
+                //
+                // Note that currently, the idle CPU selection (pick_idle_cpu)
+                // is not optimized for this kind of architecture, where big
+                // and LITTLE cores are in different node/LLCs.
+                'outer: for (_dist, ncpdoms) in v.neighbor_map.borrow().iter() {
+                    for ncpdom_id in ncpdoms.borrow().iter() {
+                        if let Some(is_big) = cpdom_types.get(ncpdom_id) {
+                            if *is_big == key.is_big {
+                                v.cpdom_alt_id.set(*ncpdom_id);
+                                break 'outer;
+                            }
+                        }
+                    }
+                }
+            }
+        }
+
         Some(cpdom_map)
     }
 
@@ -635,37 +668,44 @@ impl<'a> Scheduler<'a> {
         // Initialize CPU capacity
         for (_, cpu) in topo.cpu_fids_performance.iter().enumerate() {
             skel.maps.rodata_data.cpu_capacity[cpu.cpu_id] = cpu.cpu_cap as u16;
+            skel.maps.rodata_data.cpu_big[cpu.cpu_id] = cpu.big_core as u8;
+            skel.maps.rodata_data.cpu_turbo[cpu.cpu_id] = cpu.turbo_core as u8;
         }
 
         // If cpu_pref_order is not specified, initialize CPU order
         // topologically sorted by a cpu, node, llc, max_freq, and core order.
         // Otherwise, follow the specified CPU preference order.
-        let mut cpu_pf_order = vec![];
-        let mut cpu_ps_order = vec![];
-        if opts.cpu_pref_order == "" {
-            for cpu in topo.cpu_fids_performance.iter() {
-                cpu_pf_order.push(cpu.cpu_id);
-            }
-            for cpu in topo.cpu_fids_powersave.iter() {
-                cpu_ps_order.push(cpu.cpu_id);
-            }
+        let (cpu_pf_order, cpu_ps_order) = if opts.cpu_pref_order.is_empty() {
+            (
+                topo.cpu_fids_performance
+                    .iter()
+                    .map(|cpu| cpu.cpu_id)
+                    .collect(),
+                topo.cpu_fids_powersave
+                    .iter()
+                    .map(|cpu| cpu.cpu_id)
+                    .collect(),
+            )
         } else {
             let cpu_list = read_cpulist(&opts.cpu_pref_order).unwrap();
             let pref_mask = Cpumask::from_cpulist(&opts.cpu_pref_order).unwrap();
             if pref_mask != topo.all_cpus_mask {
                 panic!("--cpu_pref_order does not cover the whole CPUs.");
             }
-            cpu_pf_order = cpu_list.clone();
-            cpu_ps_order = cpu_list.clone();
-        }
+            (cpu_list.clone(), cpu_list)
+        };
         for (pos, cpu) in cpu_pf_order.iter().enumerate() {
             skel.maps.rodata_data.cpu_order_performance[pos] = *cpu as u16;
         }
         for (pos, cpu) in cpu_ps_order.iter().enumerate() {
             skel.maps.rodata_data.cpu_order_powersave[pos] = *cpu as u16;
         }
-        info!("CPU pref order in performance mode: {:?}", cpu_pf_order);
-        info!("CPU pref order in powersave mode: {:?}", cpu_ps_order);
+        if !opts.powersave {
+            info!("CPU pref order in performance mode: {:?}", cpu_pf_order);
+        }
+        if !opts.performance {
+            info!("CPU pref order in powersave mode: {:?}", cpu_ps_order);
+        }
 
         // Initialize compute domain contexts
         for (k, v) in topo.cpdom_map.iter() {
@@ -709,6 +749,7 @@ impl<'a> Scheduler<'a> {
         skel.maps.rodata_data.verbose = opts.verbose;
         skel.maps.rodata_data.slice_max_ns = opts.slice_max_us * 1000;
         skel.maps.rodata_data.slice_min_ns = opts.slice_min_us * 1000;
+        skel.maps.rodata_data.preempt_shift = opts.preempt_shift;
 
         skel.struct_ops.lavd_ops_mut().flags = *compat::SCX_OPS_ALLOW_QUEUED_WAKEUP
             | *compat::SCX_OPS_ENQ_EXITING
@@ -767,7 +808,7 @@ impl<'a> Scheduler<'a> {
             nr_active: tx.nr_active,
         }) {
             Ok(()) | Err(TrySendError::Full(_)) => 0,
-            Err(e) => panic!("failed to send on intrspc_tx ({})", &e),
+            Err(e) => panic!("failed to send on intrspc_tx ({})", e),
         }
     }
 
@@ -911,7 +952,7 @@ impl<'a> Scheduler<'a> {
 
         let _ = match profile {
             PowerProfile::Performance => self.set_power_profile(LAVD_PM_PERFORMANCE),
-            PowerProfile::Balanced => self.set_power_profile(LAVD_PM_BALANCED),
+            PowerProfile::Balanced { .. } => self.set_power_profile(LAVD_PM_BALANCED),
             PowerProfile::Powersave => self.set_power_profile(LAVD_PM_POWERSAVE),
             PowerProfile::Unknown => {
                 // We don't know how to handle an unknown energy profile,
@@ -954,7 +995,7 @@ impl<'a> Scheduler<'a> {
         }
         self.rb_mgr.consume().unwrap();
 
-        self.struct_ops.take();
+        let _ = self.struct_ops.take();
         uei_report!(&self.skel, uei)
     }
 }
@@ -974,7 +1015,9 @@ fn init_log(opts: &Opts) {
         _ => simplelog::LevelFilter::Trace,
     };
     let mut lcfg = simplelog::ConfigBuilder::new();
-    lcfg.set_time_level(simplelog::LevelFilter::Error)
+    lcfg.set_time_offset_to_local()
+        .expect("Failed to set local time offset")
+        .set_time_level(simplelog::LevelFilter::Error)
         .set_location_level(simplelog::LevelFilter::Off)
         .set_target_level(simplelog::LevelFilter::Off)
         .set_thread_level(simplelog::LevelFilter::Off);
