@@ -39,12 +39,14 @@ const GSTAT_LO_FB_USAGE: usize = bpf_intf::global_stat_id_GSTAT_LO_FB_USAGE as u
 const GSTAT_FB_CPU_USAGE: usize = bpf_intf::global_stat_id_GSTAT_FB_CPU_USAGE as usize;
 const GSTAT_ANTISTALL: usize = bpf_intf::global_stat_id_GSTAT_ANTISTALL as usize;
 const GSTAT_SKIP_PREEMPT: usize = bpf_intf::global_stat_id_GSTAT_SKIP_PREEMPT as usize;
+const GSTAT_FIXUP_VTIME: usize = bpf_intf::global_stat_id_GSTAT_FIXUP_VTIME as usize;
 
 const LSTAT_SEL_LOCAL: usize = bpf_intf::layer_stat_id_LSTAT_SEL_LOCAL as usize;
 const LSTAT_ENQ_LOCAL: usize = bpf_intf::layer_stat_id_LSTAT_ENQ_LOCAL as usize;
 const LSTAT_ENQ_WAKEUP: usize = bpf_intf::layer_stat_id_LSTAT_ENQ_WAKEUP as usize;
 const LSTAT_ENQ_EXPIRE: usize = bpf_intf::layer_stat_id_LSTAT_ENQ_EXPIRE as usize;
 const LSTAT_ENQ_REENQ: usize = bpf_intf::layer_stat_id_LSTAT_ENQ_REENQ as usize;
+const LSTAT_ENQ_DSQ: usize = bpf_intf::layer_stat_id_LSTAT_ENQ_DSQ as usize;
 const LSTAT_MIN_EXEC: usize = bpf_intf::layer_stat_id_LSTAT_MIN_EXEC as usize;
 const LSTAT_MIN_EXEC_NS: usize = bpf_intf::layer_stat_id_LSTAT_MIN_EXEC_NS as usize;
 const LSTAT_OPEN_IDLE: usize = bpf_intf::layer_stat_id_LSTAT_OPEN_IDLE as usize;
@@ -133,6 +135,8 @@ pub struct LayerStats {
     pub enq_expire: f64,
     #[stat(desc = "% re-enqueued due to RT preemption")]
     pub enq_reenq: f64,
+    #[stat(desc = "% enqueued into the layer's LLC DSQs")]
+    pub enq_dsq: f64,
     #[stat(desc = "count of times exec duration < min_exec_us")]
     pub min_exec: f64,
     #[stat(desc = "total exec durations extended due to min_exec_us")]
@@ -253,6 +257,7 @@ impl LayerStats {
             enq_wakeup: lstat_pct(LSTAT_ENQ_WAKEUP),
             enq_expire: lstat_pct(LSTAT_ENQ_EXPIRE),
             enq_reenq: lstat_pct(LSTAT_ENQ_REENQ),
+            enq_dsq: lstat_pct(LSTAT_ENQ_DSQ),
             min_exec: lstat_pct(LSTAT_MIN_EXEC),
             min_exec_us: (lstat(LSTAT_MIN_EXEC_NS) / 1000) as u64,
             open_idle: lstat_pct(LSTAT_OPEN_IDLE),
@@ -320,11 +325,12 @@ impl LayerStats {
 
         writeln!(
             w,
-            "  {:<width$}  tot={:7} local_sel/enq={}/{} wake/exp/reenq={}/{}/{}",
+            "  {:<width$}  tot={:7} local_sel/enq={}/{} enq_dsq={} wake/exp/reenq={}/{}/{}",
             "",
             self.total,
             fmt_pct(self.sel_local),
             fmt_pct(self.enq_local),
+            fmt_pct(self.enq_dsq),
             fmt_pct(self.enq_wakeup),
             fmt_pct(self.enq_expire),
             fmt_pct(self.enq_reenq),
@@ -496,6 +502,8 @@ pub struct SysStats {
     pub antistall: u64,
     #[stat(desc = "Number of times preemptions of non-scx tasks were avoided")]
     pub skip_preempt: u64,
+    #[stat(desc = "Number of times vtime was out of range and fixed up")]
+    pub fixup_vtime: u64,
     #[stat(desc = "fallback CPU")]
     pub fallback_cpu: u32,
     #[stat(desc = "per-layer statistics")]
@@ -552,6 +560,7 @@ impl SysStats {
                 * 100.0,
             antistall: stats.bpf_stats.gstats[GSTAT_ANTISTALL],
             skip_preempt: stats.bpf_stats.gstats[GSTAT_SKIP_PREEMPT],
+            fixup_vtime: stats.bpf_stats.gstats[GSTAT_FIXUP_VTIME],
             fallback_cpu: fallback_cpu as u32,
             fallback_cpu_util: stats.bpf_stats.gstats[GSTAT_FB_CPU_USAGE] as f64
                 / elapsed_ns as f64
@@ -563,7 +572,7 @@ impl SysStats {
     pub fn format<W: Write>(&self, w: &mut W) -> Result<()> {
         writeln!(
             w,
-            "tot={:7} local_sel/enq={}/{} open_idle={} affn_viol={} hi/lo={}/{} skip_preempt={}",
+            "tot={:7} local_sel/enq={}/{} open_idle={} affn_viol={} hi/lo={}/{}",
             self.total,
             fmt_pct(self.local_sel),
             fmt_pct(self.local_enq),
@@ -571,12 +580,11 @@ impl SysStats {
             fmt_pct(self.affn_viol),
             fmt_pct(self.hi_fb),
             fmt_pct(self.lo_fb),
-            self.skip_preempt,
         )?;
 
         writeln!(
             w,
-            "busy={:5.1} util/hi/lo={:7.1}/{}/{} fallback_cpu/util={:3}/{:4.1} proc={:?}ms antistall={}",
+            "busy={:5.1} util/hi/lo={:7.1}/{}/{} fallback_cpu/util={:3}/{:4.1} proc={:?}ms",
             self.busy,
             self.util,
             fmt_pct(self.hi_fb_util),
@@ -584,7 +592,6 @@ impl SysStats {
             self.fallback_cpu,
             self.fallback_cpu_util,
             self.proc_ms,
-            self.antistall,
         )?;
 
         writeln!(
@@ -593,6 +600,12 @@ impl SysStats {
             self.excl_collision, self.excl_preempt, self.excl_idle, self.excl_wakeup
         )?;
 
+        writeln!(
+            w,
+            "skip_preempt={} antistall={} fixup_vtime={}",
+            self.skip_preempt, self.antistall, self.fixup_vtime
+        )?;
+
         Ok(())
     }
 
@@ -640,7 +653,7 @@ pub fn server_data() -> StatsServerData<StatsReq, StatsRes> {
         req_ch.send(StatsReq::Hello(tid))?;
         let mut stats = Some(match res_ch.recv()? {
             StatsRes::Hello(v) => v,
-            res => bail!("invalid response to Hello: {:?}", &res),
+            res => bail!("invalid response to Hello: {:?}", res),
         });
 
         let read: Box<dyn StatsReader<StatsReq, StatsRes>> =
@@ -648,7 +661,7 @@ pub fn server_data() -> StatsServerData<StatsReq, StatsRes> {
                 req_ch.send(StatsReq::Refresh(tid, stats.take().unwrap()))?;
                 let (new_stats, sys_stats) = match res_ch.recv()? {
                     StatsRes::Refreshed(v) => v,
-                    res => bail!("invalid response to Refresh: {:?}", &res),
+                    res => bail!("invalid response to Refresh: {:?}", res),
                 };
                 stats = Some(new_stats);
                 sys_stats.to_json()
@@ -661,7 +674,7 @@ pub fn server_data() -> StatsServerData<StatsReq, StatsRes> {
         req_ch.send(StatsReq::Bye(current().id())).unwrap();
         match res_ch.recv().unwrap() {
             StatsRes::Bye => {}
-            res => panic!("invalid response to Bye: {:?}", &res),
+            res => panic!("invalid response to Bye: {:?}", res),
         }
     });
 
