@@ -26,14 +26,19 @@ enum consts {
 	MAX_LLCS		= 64,
 	MAX_DSQS_PER_LLC	= 8,
 	MAX_TASK_PRIO		= 39,
+	MAX_TOPO_NODES		= 1024,
 
 	NSEC_PER_USEC		= 1000ULL,
 	NSEC_PER_MSEC		= (1000ULL * NSEC_PER_USEC),
 	MSEC_PER_SEC		= 1000ULL,
 	NSEC_PER_SEC		= NSEC_PER_MSEC * MSEC_PER_SEC,
 
+	LOAD_BALANCE_SLACK	= 20ULL,
+
 	// kernel definitions
 	CLOCK_BOOTTIME		= 7,
+
+	STATIC_ALLOC_PAGES_GRANULARITY = 8,
 };
 
 enum p2dq_timers_defs {
