@@ -13,9 +13,12 @@ fn main() {
         .add_source("src/bpf/deadline.bpf.c")
         .add_source("src/bpf/placement.bpf.c")
         .add_source("src/bpf/deadline.bpf.c")
+        .add_source("../../../lib/arena.bpf.c")
+        .add_source("../../../lib/bitmap.bpf.c")
         .add_source("../../../lib/cpumask.bpf.c")
         .add_source("../../../lib/sdt_task.bpf.c")
         .add_source("../../../lib/sdt_alloc.bpf.c")
+        .add_source("../../../lib/topology.bpf.c")
         .compile_link_gen()
         .unwrap();
 }
