@@ -4,13 +4,9 @@
 // GNU General Public License version 2.
 
 fn main() {
-    println!("cargo::rerun-if-changed=src/protos/perfetto_scx.proto");
-    protobuf_codegen::Codegen::new()
-        .pure()
-        .cargo_out_dir("protos_gen/")
-        .input("src/protos/perfetto_scx.proto")
-        .include("src/protos")
-        .run_from_script();
+    prost_build::Config::new()
+        .compile_protos(&["src/protos/perfetto_scx.proto"], &["src/protos"])
+        .unwrap();
     scx_utils::BpfBuilder::new()
         .unwrap()
         .enable_intf("src/bpf/intf.h", "bpf_intf.rs")
